class Class
	# Utility for defining 'classial' enum type
	module Enum
		VERSION = '0.0.2'.freeze
		
		def self.included(base)
			base.send :extend, ClassMethods
			base.private_class_method :new
		end
		
		module ClassMethods
			# @!method field(name, *args, **kwargs, &block)
			#   @param [#to_sym] name
			#   @param args Arguments for constructor
			#   @param kwargs Keyword arguments for constructor
			#     **This argument exists only if ruby 2.0+**
			#   @param block A block for constructor
			#   @return A created object
			if RUBY_VERSION >= '2.0'
				eval <<-RUBY
					# @param [#to_sym] name
					# @param args Arguments for constructor
					# @param kwargs Keyword arguments for constructor
					# @param block A block for constructor
					# @return A created object
					def field(name, *args, **kwargs, &block)
						if kwargs.empty?
							obj = self.send(:new, *args, &block)
						else
							obj = self.send(:new, *args, **kwargs, &block)
						end
						name = name.to_sym
						const_set(name, obj)
						@enum_fields ||= {}
						@enum_fields[name] = obj
						obj
					end
				RUBY
			else
				eval <<-RUBY
					# @param [#to_sym] name
					# @param args Arguments for constructor
					# @param block A block for constructor
					# @return A created object
					def field(name, *args, &block)
						obj = self.send(:new, *args, &block)
						name = name.to_sym
						const_set(name, obj)
						@enum_fields ||= {}
						@enum_fields[name] = obj
						obj
					end
				RUBY
			end
			
			# Checks a field of given name exists
			# You can also do like `Animal::DOG.instance_of?(Animal)`
			# 
			# @param [#to_sym] name
			# @return [Boolean]
			def key?(name)
				name = name.to_sym
				@enum_fields.key?(name)
			end
			
			# Checks given value is a field of this enum
			# 
			# @param obj
			# @return [Boolean]
			def value?(obj)
				@enum_fields.value?(obj)
			end
			
			# Gets a field by given name
			# You can also do like `Animal::DOG`
			# 
			# @param [#to_sym] name
			# @return A enum field with given name. If not exists, nil
			def value(name)
				@enum_fields[name.to_sym]
			end
			
			# Gets an array of enum fields
			# @return [Array] (copied) An array of enum fields
			def values
				@enum_fields.values
			end
			alias_method :to_a, :values
			
			# Gets an array of enum field names
			# @return [Array<Symbol>] (copied) An array of enum field names
			def keys
				@enum_fields.keys
			end
			
			# Gets an hash of enum fields
			# @return [Hash{Symbol, Object}] (copied) An hash of enum fields
			def to_h
				@enum_fields.dup
			end
			
			# @overload fetch(name)
			#   Returns a enum field for the given name.
			#   If the name can't be found, it will raise an {KeyError} exception
			#   @param [#to_sym] name
			#   @raise [KeyError] If the name can't be found
			# @overload fetch(name, default)
			#   Returns a enum field for the given name.
			#   If the name can't be found, the given *default* will be returned
			#   @param [#to_sym] name
			#   @param default
			def fetch(*args)
				case args.length
				when 1
					name = args[0]
					if key?(name)
						value(name)
					else
						raise KeyError, "key not found: #{name.inspect}"
					end
				when 2
					name = args[0]
					default = args[1]
					if key?(name)
						value(name)
					else
						default
					end
				else
					raise ArgumentError, "Invalid arguments length"
				end
			end
		end
	end
end
